use std::{
    array,
    collections::HashMap,
    io::{BufReader, BufWriter, Write},
    panic,
    path::PathBuf,
    sync::{
        mpsc::{self},
        Arc, Mutex,
    },
    thread,
    time::{Duration, Instant},
};

use arrayvec::ArrayVec;
use colored::Colorize;
pub use config::Config;
use demux::Reference;
use flate2::{read::MultiGzDecoder, write::GzEncoder, Compression};
use from_unstable::{self as fun, TryMap};
use indexmap::{map::Entry, IndexMap};
use io::{Context, ContextualizedReader, ContextualizedWriter};
use num_format::{Locale, ToFormattedString};
use parse::{
    fastq,
    samples::{self, Samples},
};
use priority_queue::PriorityQueue;
use short::ful;
use snafu::{ensure, ResultExt, Snafu};

use crate::fastq::Block;

pub mod config;
mod demux;
mod dispatcher;
mod from_unstable;
mod io;
mod parse;
mod reader;
mod short;
mod writer;

/// Standard input files ordering.
pub(crate) const INPUT_INFIXES: [&str; 4] = ["I1", "I2", "R1", "R2"];

/// Standard output files ordering.
pub(crate) const OUTPUT_INFIXES: [&str; 2] = ["R1", "R2"];

/// Standard module letters.
pub(crate) const MODULE_LETTERS: &[u8; 4] = b"ABCD";

/// One barcode is 4 modules (watch goofy letters ordering!).
type Barcode = [usize; 4]; // [A, C, B, D]

/// Sample identifier may be undetermined.
type SampleCode = Option<usize>;

#[allow(clippy::too_many_lines)] // Ongoing experiment.
pub fn demultiplex(cf: &Config) -> Result<(), Error> {
    let tic = Instant::now();

    //==============================================================================================
    println!("Spawn input readers.");
    // (do this first so they can start while we parse modules and samples)

    // Count input allocations creation/recycling for performance statistics.
    let in_allocs_created = Arc::new(Mutex::new(0));
    let in_allocs_reused = Arc::new(Mutex::new(0));

    // Spawn one reader thread per input (infix) file.
    let mut readers: [_; INPUT_INFIXES.len()] = fun::try_from_fn(|i| -> Result<_, Error> {
        let infix = INPUT_INFIXES[i];
        let path = &cf.inputs[i];
        println!("  {}", path.display().to_string().blue());

        // Setup disk-read | unzip | parse stream.
        let file = io::open_file(path)?;
        let gz = MultiGzDecoder::new(file);
        let reader = BufReader::new(gz);
        let reader = ContextualizedReader::new(path.clone(), reader);
        let reader = fastq::BlocksReader::new(reader);

        // Spawn thread.
        let rcc = cf.readers_channel_capacity;
        let (block_sender, block_receiver) = mpsc::sync_channel::<Option<Block>>(rcc);
        let (alloc_sender, alloc_receiver) = mpsc::sync_channel::<Block>(rcc);
        let name = format!("{infix}dmoxreader");
        let thread = reader::Inner {
            reader,
            input: alloc_receiver,
            output: block_sender,
            new_allocs: in_allocs_created.clone(),
            recycled_allocs: in_allocs_reused.clone(),
        };
        let handle = thread::Builder::new()
            .name(name.clone())
            .spawn(move || -> Result<(), reader::Error> { thread.run() })
            .with_context(|_| SpawnErr { name })?;
        Ok(reader::Outer {
            infix,
            path: path.clone(),
            handle,
            input: alloc_sender,
            output: block_receiver,
        })
    })?;

    //----------------------------------------------------------------------------------------------
    println!("Read all expected modules from references.");
    let [mut ref_a, mut ref_b, mut ref_c, mut ref_d] = // (mutable because of their internal caches)
        fun::try_from_fn(|i| -> Result<_, Error> {
            let letter = MODULE_LETTERS[i];
            let path = &cf.references[i];
            println!("  {}", path.display().to_string().blue());
            let reader = ContextualizedReader::from_file(path)?;
            let cf = cf.into();
            let modules = parse::modules(reader, letter, &cf)
                .with_context(|_| ModulesErr { letter, path })?;
            Ok(Reference::new(modules, cf))
        })?;

    //----------------------------------------------------------------------------------------------
    println!("Read samples from schema.");
    let path = &cf.schema;
    println!("  {}", path.display().to_string().blue());
    let reader = ContextualizedReader::from_file(path)?;
    let samples = Samples::parse(reader)?;

    //----------------------------------------------------------------------------------------------
    println!("Open table stream.");
    // (do it now to error early if it can't be opened, but actual writing there happens later)
    let path = &cf.barcodes_table_file;
    let header = "Barcode\tTotal\tCorrect_Reads\tCorrected_Reads\n";
    let table_file = {
        println!("  {}", path.display().to_string().blue());
        let mut writer = ContextualizedWriter::from_file(path)?;
        writer
            .write_all(header.as_bytes())
            .with_context(|_| WriteLineErr { context: &writer })?;
        writer.bump_line();
        writer
    };
    let mut counters = IndexMap::new();

    //----------------------------------------------------------------------------------------------
    // There are two outputs files per sample + the undetermined files,
    // and compressing/writing seems to be the CPU bottleneck here.
    // For this reason: prepare one writer for every output file,
    // dedicated to compress/write output in its own file,
    // then spawn as many threads as required to parallelize these.

    println!("Prepare samples writers.");
    println!(
        "  {}",
        (cf.sample)("<sample_id>", "{I1,I2,R1,R2}")
            .display()
            .to_string()
            .blue()
    );

    // Buffers are filled by this main thread, then consumed by writer threads.
    let mut output_buffers = dispatcher::BufferMap::new();
    // Writers are passed to writer threads and consumed by them on termination.
    let mut all_writers = HashMap::new(); // { sample code ↦  [R1 writer, R2 writer] }

    // Allocations with blocks are sent to writers
    // which send them back for recycling once they are done with them.
    let (writer_output_sender, writer_output_receiver) = mpsc::channel::<writer::Output<_>>();
    // Monitor to get a sense whether this optimisation is useful.
    let mut out_allocs_created = 0;

    let new_output_writer = |path: PathBuf| -> Result<_, Error> {
        // Create writer with zip | disk IO stream.
        let file = io::create_file(&path)?;
        let writer = BufWriter::with_capacity(cf.writers_capacity, file);
        let gz = GzEncoder::new(writer, Compression::new(cf.output_compression_level));
        let writer = ContextualizedWriter::new(path, gz);
        Ok(writer)
    };
    let mut new_output_counter = || {
        (
            OUTPUT_INFIXES.map(|_| {
                out_allocs_created += 1;
                Vec::new()
            }),
            0,
        )
    };
    for (&code, sample_id) in &samples.map {
        all_writers.insert(
            Some(code),
            #[allow(unstable_name_collisions)] // Purposedly downgrading from nightly.
            OUTPUT_INFIXES
                .try_map(|infix| new_output_writer((cf.sample)(&ful(sample_id), infix)))?,
        );
        output_buffers.insert(Some(code), new_output_counter());
    }
    // + One for undetermined samples.
    if let Some(undetermined) = &cf.undetermined {
        println!("Open undetermined output files.");
        println!("  {}", undetermined("{R1,R2}").display().to_string().blue());
        all_writers.insert(
            None,
            #[allow(unstable_name_collisions)] // Purposedly downgrading from nightly.
            OUTPUT_INFIXES.try_map(|infix| new_output_writer(undetermined(infix)))?,
        );
        output_buffers.insert(None, new_output_counter());
    }

    //----------------------------------------------------------------------------------------------
    println!("Spawn {} writer threads.", cf.n_writers);
    let workers = (0..cf.n_writers)
        .map(|id| -> Result<_, Error> {
            // Spawn thread.
            let (input_sender, input_receiver) = mpsc::channel::<writer::Input<_>>();
            let name = format!("{id:02}-dmoxwriter");
            let output_sender = writer_output_sender.clone();
            let thread = writer::Inner {
                id,
                input: input_receiver,
                output: output_sender,
            };
            // Spawn writer thread.
            let handle = thread::Builder::new()
                .name(name.clone())
                .spawn(move || -> Result<(), writer::Error> { thread.run() })
                .with_context(|_| SpawnErr { name })?;
            Ok(writer::Outer {
                handle,
                input: input_sender,
            })
        })
        .collect::<Result<Vec<_>, _>>()?;

    //----------------------------------------------------------------------------------------------
    println!("Spawn dispatcher thread.");
    // Keep track of idle writers and of the longest buffers to be written yet
    // (in terms of number of fastq blocks).
    let queue: PriorityQueue<_, _> = all_writers.keys().map(|k| (*k, 0)).collect();
    let queue = Arc::new(Mutex::new(queue));
    let output_buffers = Arc::new(Mutex::new(output_buffers));
    let (dispatcher_channel, termination) = mpsc::channel::<()>(); // Termination signal.
    let name = "dispatcher";
    let thread = dispatcher::Inner {
        workers,
        all_writers,
        buffers: output_buffers.clone(),
        queue: queue.clone(),
        writers_output: writer_output_receiver,
        termination,
        allocs_created: out_allocs_created,
        allocs_reused: 0,
    };
    let dispatcher = thread::Builder::new()
        .name(name.into())
        .spawn(move || -> Result<_, dispatcher::Error> { thread.run() })
        .with_context(|_| SpawnErr { name })?;

    //==============================================================================================
    println!("Process..");
    loop {
        // Extract the next fastq blocks from all input files,
        // or check that all EOF happen together.
        let mut eof = None; // Define on first information then raise on first EOF.
        let mut block_line = None; // Define on first information, should be the same for all.
        let mut hung = None; // Raise if a reader has unexpectedly stopped.
        let mut blocks = ArrayVec::<_, 4>::new();
        for (i, reader) in readers.iter_mut().enumerate() {
            // Wait on the reader to send input.
            let Ok(block) = reader.output.recv() else {
                hung = Some(i);
                break;
            };
            block_line = block.as_ref().map(|b| b.i_line);
            match (eof, &block) {
                (None, _) => {
                    eof = Some(block.is_none());
                }
                (Some(true), None) | (Some(false), Some(_)) => {}
                _ => {
                    return DissimilarFilesErr {
                        context: Context {
                            path: reader.path.clone(),
                            i_line: block_line.unwrap(),
                        },
                    }
                    .fail();
                }
            }
            blocks.push(block);
        }
        if let Some(i) = hung {
            let reader = readers.into_iter().nth(i).unwrap();
            let infix = reader.infix;
            // The reader has hung up.
            match reader.handle.join() {
                Ok(Err(e)) => return Err(e.into()), // Propagate reader error.
                Ok(Ok(())) => panic!("Unexpected correct termination of writer thread."),
                Err(e) => {
                    eprintln!(
                        "Reader thread {infix} unexpectedly exited: \
                         propagate panic to main thread.",
                    );
                    panic::resume_unwind(e);
                }
            }
        }
        if eof.unwrap() {
            println!("Reached end of all input files.");
            break;
        }
        // Got our input blocks.
        let [i1, i2, r1, r2] = array::from_fn(|i| (blocks[i].take().unwrap(), readers[i].infix));

        //------------------------------------------------------------------------------------------
        // Consistency checks.
        let context = |filename| Context {
            path: filename,
            i_line: block_line.unwrap(),
        };
        for (&(ref i, file_i), &(ref r, file_r)) in [(&i1, &r1), (&i2, &r2)] {
            ensure!(
                i.id() == r.id(),
                InconsistentIdErr {
                    context_i: context(file_i.into()),
                    context_r: context(file_r.into()),
                    id_i: ful(i.id()),
                    id_r: ful(r.id()),
                }
            );
            ensure!(
                i.read.len() == 2 * cf.module_size + 2,
                InvalidIndexErr {
                    module_size: cf.module_size,
                    act: i.read.len(),
                    seq: ful(i.read()),
                    context: context(file_i.into()),
                }
            );
        }

        //------------------------------------------------------------------------------------------
        // Find the corresponding matches against each reference to construct the barcode.
        let mut bx = [0, 0, 0, 0]; // [{A, C}, {B, D}] <> [{I1}, {I2}]
        let mut all_clear = true; // Raise if all closest matches were unique.
        let mut all_exact = true; // Raise if all match distances were null.
        let mut sample_code = None; // Fill if the sample letter has unique closest match.
        let (m1, m2) = bx.split_at_mut(2);
        for ((i, _), (u, v), matches) in [
            (&i1, ((b'A', &mut ref_a), (b'C', &mut ref_c)), m1),
            (&i2, ((b'B', &mut ref_b), (b'D', &mut ref_d)), m2),
        ] {
            // Match index against the reference.
            // (watch the reversion between q1 and q2 here)
            let (q1, q2) = i.read().split_at(cf.module_size);
            let q2 = &q2[1..]; // Remove (and ignore) stitch base.
            for ((q, (letter, reference)), m) in [(q2, u), (q1, v)].iter_mut().zip(matches) {
                let Some((code, distance)) = reference.closest(q) else {
                    all_clear = false;
                    continue;
                };
                if let Some(max) = cf.max_distance {
                    if distance > max {
                        all_clear = false;
                        continue;
                    }
                }
                if distance > 0 {
                    all_exact = false;
                }
                *m = code + 1;
                if *letter == samples.letter {
                    sample_code = Some(*m);
                }
            }
        }

        // Update counters,
        // considering unclear sample as neither 'exact' or 'corrected'.
        let [exact, corrected] =
            [all_exact, !all_exact].map(|b| if all_clear { u64::from(b) } else { 0 });
        match counters.entry(bx) {
            Entry::Vacant(entry) => {
                entry.insert(Counters {
                    total: 1,
                    exact,
                    corrected,
                });
            }
            Entry::Occupied(mut entry) => {
                let counters = entry.get_mut();
                counters.total += 1;
                counters.exact += exact;
                counters.corrected += corrected;
            }
        }

        //------------------------------------------------------------------------------------------
        // Construct output fastq blocks.
        let [[i1, i2], [r1, r2]] = if let (None, None) = (&cf.undetermined, sample_code) {
            // Unclear sample code and no dedicated output file: skip.
            [[i1, i2], [r1, r2]].map(|[(a, _), (b, _)]| [a, b])
        } else {
            if let Some(code) = sample_code {
                ensure!(
                    samples.map.contains_key(&code),
                    NoSampleErr {
                        letter: samples.letter,
                        code,
                        context: context(i1.1.into()),
                    }
                );
            }

            // Then for both reads..
            let [[i1, i2], r1r2] = [[i1, i2], [r1, r2]].map(|[(a, _), (b, _)]| [a, b]);
            // Pick the right output buffers.
            let mut buffers_lock = output_buffers.lock().expect("poisoned buffers?");
            let mut queue_lock = queue.lock().expect("poisoned queue?");
            let (buffers, n_blocks) = buffers_lock.get_mut(&sample_code).unwrap();
            // Push fastq blocks inside.
            for (r, buffer) in r1r2.iter().zip(buffers) {
                macro_rules! w {
                    ($content:expr) => {
                        buffer.write_all($content).unwrap();
                    };
                }
                // Identifier head.
                w!(r.id_head());
                if cf.output_bx {
                    w!(b" BX:Z:");
                    write_code(buffer, &bx).unwrap();
                }
                if cf.output_rx {
                    w!(b" RX:Z:");
                    w!(i1.read());
                    w!(b"+");
                    w!(i2.read());
                }
                if cf.output_qx {
                    w!(b" QX:Z:");
                    w!(i1.quality());
                    w!(b"+");
                    w!(i2.quality());
                }
                if cf.output_id_tail {
                    w!(b" ");
                    w!(r.id_tail());
                }

                w!(b"\n");

                // Actual Sequence.
                w!(r.read_quality());
            }
            *n_blocks += 1;
            // Tell the dispatcher it's ready.
            queue_lock.change_priority(&sample_code, *n_blocks);
            drop(queue_lock);
            drop(buffers_lock);
            [[i1, i2], r1r2]
        };

        //------------------------------------------------------------------------------------------

        // Done with these read allocations: send them back for recycling.
        for (reader, block) in readers.iter_mut().zip([i1, i2, r1, r2]) {
            let Ok(()) = reader.input.send(block) else {
                continue; // No recycling on hang up: either the reader errored or eof.
            };
        }

        // Don't resume and ask for new reads if the queue is too much filled.
        // This sets up backpressure on reader threads,
        // and prevents the RAM from getting filled with input.
        loop {
            let n_blocks = {
                let (_, &n) = queue.lock().expect("poisoned queue?").peek().unwrap();
                n
            };
            if n_blocks < cf.max_queued_blocks {
                break;
            }
            // .. wait for it to empty instead.
            thread::sleep(Duration::from_millis(1));
        }
    }

    //==============================================================================================
    println!("Wait for readers to terminate.");
    for reader in readers {
        drop(reader.input);
        let infix = reader.infix;
        match reader.handle.join() {
            Ok(Ok(())) => {}
            Ok(Err(err)) => return Err(err.into()),
            Err(err) => {
                eprintln!("Reader thread {infix} unexpectedly exited: propagate panic.",);
                panic::resume_unwind(err);
            }
        }
    }

    println!("Send termination signal to dispatcher.");
    dispatcher_channel.send(()).map_err(|e| Error::Message {
        target: "dispatcher",
        erm: e.to_string(),
    })?;

    println!("Wait for dispatcher to terminate.");
    let (out_allocs_created, out_allocs_reused) = match dispatcher.join() {
        Ok(Ok(pair)) => pair,
        Ok(Err(err)) => return Err(err)?,
        Err(err) => {
            eprintln!("Dispatcher thread unexpectedly exited: propagate panic.");
            panic::resume_unwind(err);
        }
    };

    // Sanity checks.
    for (buffers, n_blocks) in output_buffers.lock().expect("buffers poisoned?").values() {
        for buffer in buffers {
            assert!(
                *n_blocks == 0 && buffer.is_empty(),
                "No unwritten data should remain."
            );
        }
    }

    //----------------------------------------------------------------------------------------------
    // Display allocation reuse statistics: useful to determine input/output bind.
    let fmt = |n: u64| n.to_formatted_string(&Locale::en);
    let e = "All threads concurrent for this lock should have terminated.";
    let [ic, ir, oc, or] = [
        *in_allocs_created.lock().expect(e),
        *in_allocs_reused.lock().expect(e),
        out_allocs_created,
        out_allocs_reused,
    ]
    .map(fmt);
    let (hc, hr) = ("allocated", "reused");
    let max = |a: &[&str]| a.iter().map(|s| s.len()).max().unwrap();
    let (wc, wr) = (max(&[&ic, &oc, hc]), max(&[&ir, &or, hr]));
    println!("Fastq blocks allocations summary:");
    println!("          {hc:>wc$} | {hr:>wr$}");
    println!("   input: {ic:>wc$} | {ir:>wr$}");
    println!("  output: {oc:>wc$} | {or:>wr$}");

    // Display query cache statistics.
    let (hm, hq, hs) = ("cache misses", "queries", "cache hits");
    let [([ha, qa], sa), ([hb, qb], sb), ([hc, qc], sc), ([hd, qd], sd)] =
        [ref_a, ref_b, ref_c, ref_d].map(|r| {
            let (m, q) = (r.n_queries - r.n_hits, r.n_queries);
            #[allow(clippy::cast_precision_loss)]
            (
                [m, q].map(fmt),
                format!("{:0.3}%", 100. * 1. - m as f64 / q as f64),
            )
        });
    let (wm, wq, ws) = (
        max(&[&ha, &hb, &hc, &hd, hm]),
        max(&[&qa, &qb, &qc, &qd, hq]),
        max(&[&sa, &sb, &sc, &sd, hs]),
    );
    println!("Actual distance calculations to match indexes against reference modules:");
    println!("             {hm:>wm$} / {hq:>wq$} → {hs:>ws$}");
    println!("  against A: {ha:>wm$} / {qa:>wq$} → {sa:>ws$}");
    println!("  against B: {hb:>wm$} / {qb:>wq$} → {sb:>ws$}");
    println!("  against C: {hc:>wm$} / {qc:>wq$} → {sc:>ws$}");
    println!("  against D: {hd:>wm$} / {qd:>wq$} → {sd:>ws$}");

    //----------------------------------------------------------------------------------------------
    println!("Writing clear/unclear files.");
    let mut writer = table_file;
    for (code, counters) in counters {
        let Counters {
            total,
            exact,
            corrected,
        } = counters;
        |w: &mut ContextualizedWriter<_>| -> Result<(), _> {
            write_code(w, &code)?;
            writeln!(w, "\t{total}\t{exact}\t{corrected}")?;
            w.bump_line();
            Ok(())
        }(&mut writer)
        .with_context(|_| WriteLineErr { context: &writer })?;
    }

    let elapsed = tic.elapsed();
    println!("Done in {:0.5}s.", elapsed.as_secs_f64());
    Ok(())
}

/// For every barcode.
struct Counters {
    /// How many times it has been produced.
    total: u64,
    /// How many time was it an exact match against all references
    /// (impossible if it contains `00`).
    exact: u64,
    /// How many it has been corrected to match against the references
    /// (non-null distance) (also impossible if it contains `00`).
    corrected: u64,
}

fn write_code(writer: &mut impl Write, bx: &Barcode) -> Result<(), std::io::Error> {
    for (letter, code) in "ACBD".chars().zip(bx) {
        write!(writer, "{letter}{code:02}")?;
    }
    Ok(())
}

//==================================================================================================
#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum Error {
    #[snafu(display("When loading barcode modules {letter:?} from {}:\n{source}", path.display()))]
    Modules {
        letter: char,
        path: PathBuf,
        source: parse::modules::Error,
    },
    #[snafu(transparent)]
    Io { source: io::Error },
    #[snafu(display("{context}Reached EOF before the other(s)."))]
    DissimilarFiles { context: Context },
    #[snafu(display(
        "Read ids differ between files:\n\
         {context_i}{id_i:?}\n{context_r}{id_r:?}"
    ))]
    InconsistentId {
        context_i: Context,
        context_r: Context,
        id_i: String,
        id_r: String,
    },
    #[snafu(display("{context}Invalid index size:\n\
                     Expected 2 * {module_size} + 1 = {}, got instead: {act}:\n{seq:?}",
                     2 * module_size + 1))]
    InvalidIndex {
        module_size: usize,
        act: usize,
        seq: String,
        context: Context,
    },
    #[snafu(transparent)]
    Samples { source: samples::Error },
    #[snafu(display("{context}No sample is associated to code {letter}{code:02}."))]
    NoSample {
        letter: char,
        code: usize,
        context: Context,
    },
    #[snafu(display("Could not spawn thread {name:?}:\n{source}"))]
    Spawn {
        name: String,
        source: std::io::Error,
    },
    #[snafu(transparent)]
    ReaderThread { source: reader::Error },
    #[snafu(transparent)]
    DispatcherThread { source: dispatcher::Error },
    #[snafu(display("Message could not be sent from main thread to {target} thread:\n{erm}"))]
    Message {
        target: &'static str,
        erm: String, // Rendered underlying error message.
    },
    #[snafu(display("{context}IO error when writing line:{source}"))]
    WriteLine {
        source: std::io::Error,
        context: Context,
    },
}
